import numpy as np
import pandas as pd
import pandas.api.types
from lifelines.utils import concordance_index


class ParticipantVisibleError(Exception):
    pass


def score(
    solution: pd.DataFrame, submission: pd.DataFrame, row_id_column_name: str = "id"
) -> float:
    """
    >>> import pandas as pd
    >>> row_id_column_name = "id"
    >>> y_pred = {'prediction': {0: 1.0, 1: 0.0, 2: 1.0}}
    >>> y_pred = pd.DataFrame(y_pred)
    >>> y_pred.insert(0, row_id_column_name, range(len(y_pred)))
    >>> y_true = { 'efs': {0: 1.0, 1: 0.0, 2: 0.0}, 'efs_time': {0: 25.1234,1: 250.1234,2: 2500.1234}, 'race_group': {0: 'race_group_1', 1: 'race_group_1', 2: 'race_group_1'}}
    >>> y_true = pd.DataFrame(y_true)
    >>> y_true.insert(0, row_id_column_name, range(len(y_true)))
    >>> score(y_true.copy(), y_pred.copy(), row_id_column_name)
    0.75
    """
    # added --
    solution = solution.copy()
    submission = submission.copy()
    # -- added
    del solution[row_id_column_name]
    del submission[row_id_column_name]

    event_label = "efs"
    interval_label = "efs_time"
    prediction_label = "prediction"
    for col in submission.columns:
        if not pandas.api.types.is_numeric_dtype(submission[col]):
            raise ParticipantVisibleError(f"Submission column {col} must be a number")
    # Merging solution and submission dfs on ID
    merged_df = pd.concat([solution, submission], axis=1)
    merged_df.reset_index(inplace=True)
    merged_df_race_dict = dict(merged_df.groupby(["race_group"]).groups)
    metric_list = []
    for race in merged_df_race_dict.keys():
        # Retrieving values from y_test based on index
        indices = sorted(merged_df_race_dict[race])
        merged_df_race = merged_df.iloc[indices]
        # Calculate the concordance index
        c_index_race = concordance_index(
            merged_df_race[interval_label],
            -merged_df_race[prediction_label],
            merged_df_race[event_label],
        )
        metric_list.append(c_index_race)
    return float(np.mean(metric_list) - np.sqrt(np.var(metric_list)))



import ast
import collections
from collections.abc import Iterable
from typing import Iterator, Optional, Sequence

import numpy as np


def get_random_indices(n_sample: int, size_data: int) -> np.ndarray:
    """Draw `n_sample` indices from range(`size_data`)"""
    return np.random.choice(
        range(size_data),
        n_sample if n_sample < size_data else size_data,
        replace=False,
    )


def evaluate_string(x: object):
    try:
        return ast.literal_eval(str(x))
    except ValueError:
        return x


def chunks(sequence: Sequence, n: int) -> Iterator:
    """Yield successive `n`-sized chunks from `sequence`."""
    if n == -1:
        n = len(sequence)
    for i in range(0, len(sequence), n):
        yield sequence[slice(i, i + n)]


def _assert_same_type(variable, type_from_variable):
    """Check if `variable` shares type with `type_from_variable`"""
    if isinstance(variable, collections.abc.Iterable):
        if not isinstance(type_from_variable, collections.abc.Iterable):
            raise TypeError(
                f"Variable {variable=} is iterable but type is not {type_from_variable=}"
            )
        for v, typ in zip(variable, type_from_variable):
            _assert_same_type(v, typ)
    else:
        if not isinstance(variable, type(type_from_variable)):
            raise TypeError(
                f"Variable {variable=} different type than expected {type_from_variable=}"
            )


def assert_same_type_as(
    variable: object, type_from_variable: object, alternative: Optional[object] = None
):
    """Check if `variable` is same type as `type_from_variable` unless `variable` is `alternative`"""
    if variable is alternative:
        return
    try:
        _assert_same_type(variable, type_from_variable)
    except TypeError:
        raise TypeError(
            f"Variable {variable} not of expected type {type_from_variable}"
        )


def all_same_type(variable_list: Iterable, type_: type):
    for var in variable_list:
        if not isinstance(var, type_):
            raise ValueError(f"{var} not of type {type_.__name__}!")


def assert_shape(
    variable, shape: tuple, name: Optional[str] = None, ignore_none: bool = True
):
    if ignore_none:
        if variable is None:
            return
    if np.shape(variable) != shape:
        raise TypeError(
            f"{name+': ' if name is not None else ''}{variable=} doesn't have required {shape=}!"
        )


def assert_all_same_type(variable_list: Iterable, type_: type):
    for var in variable_list:
        if not isinstance(var, type_):
            raise ValueError(f"{var} not of type {type_.__name__}!")


def validate_array(array: np.ndarray, type_: str = "float", name: str | None = None):
    """Validate numpy based on `type`, e.g. no nan-values"""
    if type_ == "float":
        if np.sum(np.isnan(array.astype(float))) > 0:
            raise ValueError(
                f"Found nan-values {f'in {name}' if name is not None else ''}!"
            )
    else:
        raise ValueError(f"{type_=} unknown!")


def override_class_method(class_instance, method_name, target_class):
    class_method = getattr(target_class, method_name)

    def new_method(*args, **kwargs):
        return class_method(class_instance, *args, **kwargs)

    setattr(class_instance, method_name, new_method)



import pathlib
import warnings


def _import_torch(file):
    try:
        import torch as torch  # noqa: E501
    except ModuleNotFoundError:
        warnings.warn(
            f"Need to install `torch` to use all functionality in {pathlib.Path(file).parent}."
        )
    else:
        return torch



import datetime
import functools
import re

import numpy as np
import pandas as pd


def round_to_base(x: float, base: float):
    return base * round(x / base)


def round_offset(values: np.ndarray, decimal: int, offset: float) -> np.ndarray:
    values += offset
    values = np.round(values, decimals=decimal)
    values -= offset
    return values


def round_numpy_time_to_base_minutes(
    time: np.datetime64, base: int = 5
) -> np.datetime64:
    tm = datetime.datetime.utcfromtimestamp(time.tolist() / 1e9)
    tm += datetime.timedelta(minutes=base / 2)
    tm -= datetime.timedelta(
        minutes=tm.minute % base, seconds=tm.second, microseconds=tm.microsecond
    )
    return np.datetime64(tm)


def round_time_to_base_minutes(time: np.datetime64, base: int = 5):
    tm = pd.to_datetime(time).round(f"{base}min")
    return np.datetime64(tm)


def vetorize_time_to_base_minutes(times: np.ndarray, base: int = 5):
    function = functools.partial(round_numpy_time_to_base_minutes, base=base)
    rounded = np.array(list(map(function, times)))
    return rounded


def str_to_delta_time(string: str) -> tuple[float, str]:
    time, units, _ = re.split("([a-zA-Z]+)$", string)
    return float(time), units



import pathlib
from typing import Mapping, Optional, Sequence

import matplotlib.pyplot as plt
import numpy as np

import kego.plotting.axes_utils
import kego.plotting.figures
import kego.plotting.utils_plotting

try:
    import sklearn.metrics
except ModuleNotFoundError as e:
    raise ModuleNotFoundError(
        f"{e}\n"
        f"Optional dependencies missing, install via"
        " `poetry install --extras=train`"
    )


def plot_prediction_certainty(
    truth: np.ndarray,
    prediction_probabilities: np.ndarray,
    n_true_labels: int = 2,
    n_bins: int = 10,
    label_x: str = "True label",
    label_y: str = "Prediction probability for 'raining'",
    label_colorbar: str = "Number of Tweets",
    filename: str | pathlib.Path | None = None,
    font_size: int = 12,
    vmin=None,
    vmax=None,
    cmap="viridis",
    figure_size: tuple[float, float] | None = None,
    return_matrix: bool = False,
    axes: Optional[plt.axes] = None,
    **kwargs,
) -> tuple[plt.Axes, tuple[plt.Axes, np.ndarray]]:
    """
    Compute 2d histogram of true labels (x-axis) and
    binned prediction probabilities (y-axis)

    Parameters:
    ----------
    truth: true labels
    prediction_probabilities: probabilities for labels
    n_true_labels: number of different labels
    n_bins: number of bins for prediction probabilities
    label_x: label of x-axis
    label_y: label of y-axis
    label_colorbar: label of colorbar axis
    filename: save figure to this file, not saved if `None`
    font_size: font size of all labels of plot
    vmin: Minimum value for colorbar
    vmax: Maximum value for colorbar
    cmap: Colormap used for confusion matrix
    figure_size: size of figure [width, height]
    return_matrix: return 2d histogram matrix and axes
    axes: Matplotlib axes object

    Returns
    -------
    axes of plot
    """
    results_dic = kego.plotting.histograms.plot_histogram_2d(
        truth,
        prediction_probabilities,
        n_bins=(n_true_labels, n_bins),
        label_x=label_x,
        label_colorbar=label_colorbar,
        label_y=label_y,
        annotate=True,
        filename=filename,
        vmin=vmin,
        vmax=vmax,
        colormap=cmap,
        font_size=font_size,
        axes=axes,
        figure_size=figure_size,
        **kwargs,
    )
    if return_matrix:
        return results_dic["axes_2d_histograms_10"], results_dic["2d_histograms_10"]
    return results_dic["axes_2d_histograms_10"]


def classification_report(
    truth: Sequence,
    prediction: Sequence,
    output_dict: bool = True,
    label: str = "raining",
) -> tuple[str, Mapping]:
    """
    Compute classification report, returns precision, recall, f1-score
    and respective averages

    report returned as dict or string (see `output_dict`)
    Parameters:
    ----------
    truth: true labels
    prediction: predicted labels
    output_dict: return report as dictionary rather than string
    label: Label name of the classification

    Returns
    -------
    report
    """
    target_names = [f"not {label}", label]
    report = sklearn.metrics.classification_report(
        truth,
        prediction,
        target_names=target_names,
        output_dict=output_dict,
    )
    return report


def check_prediction(
    truth: Sequence,
    prediction: Sequence,
    output_dict: bool = True,
    filename: str | pathlib.Path | None = None,
    label: str = "raining",
    font_size: int = 14,
):
    plot_confusion_matrix(
        truth,
        prediction,
        filename=filename,
        overplot_round_base=2,
        tick_labels_x=(f"not {label}", f"{label}"),
        tick_labels_y=(f"{label}", f"not {label}"),
        font_size=font_size,
    )

    report = classification_report(
        truth,
        prediction,
        label=label,
        output_dict=output_dict,
    )
    return report


def plot_confusion_matrix(
    truth: Sequence,
    prediction: Sequence,
    filename: str | pathlib.Path | None = None,
    normalize: str = "all",
    font_size: int = 14,
    figure_size: tuple[float, float] | None = None,
    label_x: str | None = "Predicted label",
    label_y: str | None = "True label",
    vmin: float | None = None,
    vmax: float | None = None,
    text_color: str = "firebrick",
    colormap: str = "Blues",
    overplot_round_base: int | None = 2,
    tick_labels_x: tuple | None = ("not raining", "raining"),
    tick_labels_y: tuple | None = ("raining", "not raining"),
    colorbar_label: str | None = "Fraction",
    colorbar_width: float = 0.05,
    axes: plt.Axes | None = None,
    ax_colorbar: object | None = None,
):
    """
    Computes and plots confusion matrix

    Parameters:
    ----------
    truth: True labels
    prediction: Predicted labels
    filename: Save figure to this file, not saved if `None`
    normalize: Normalize style of confusion matrix
    output_dict: Return classification report as dictionary
    font_size: Font size of all labels of plot
    figure_size: Size of figure [width, height], default [6, 6]

    Returns
    -------
    Confusion matrix figure
    """
    if figure_size is None:
        figure_size = (6, 6)
    cm = sklearn.metrics.confusion_matrix(truth, prediction, normalize=normalize)
    if axes is None:
        fig, axes, ax_colorbar = kego.plotting.utils_plotting.create_axes_grid(
            1,
            1,
            figure_size=figure_size,
            unravel=True,
            colorbar_off=False,
            colorbar_width=colorbar_width,
        )
    else:
        fig = axes.get_figure()
    norm = kego.plotting.utils_plotting.get_norm("linear", vmin=vmin, vmax=vmax)
    xedges = [0, 0.5, 1]
    yedges = [0, 0.5, 1]
    X, Y = np.meshgrid(xedges, yedges)
    mesh = kego.plotting.colormesh._plot_colormesh(
        axes,
        X,
        Y,
        cm[::-1],
        norm,
        colormap,
        linewidth=0,
        rasterized=True,
    )
    kego.plotting.utils_plotting.plot_colorbar(
        mesh, cax=ax_colorbar, label=colorbar_label, font_size=font_size
    )
    kego.plotting.utils_plotting.annotate_values(
        np.array(cm),
        axes,
        2,
        2,
        color=text_color,
        round_to_base=overplot_round_base,
        font_size=font_size,
    )
    kego.plotting.utils_plotting.set_axis_tick_labels(
        axes=axes, values=[0.25, 0.75], labels=tick_labels_x, axis="x"
    )
    kego.plotting.utils_plotting.set_axis_tick_labels(
        axes=axes, values=[0.25, 0.75], labels=tick_labels_y, axis="y"
    )
    kego.plotting.axes_utils.set_axes(
        axes,
        xlim=[0, 1],
        ylim=[0, 1],
        fontsize=font_size,
        label_x=label_x,
        label_y=label_y,
    )

    kego.plotting.figures.save_figure(fig, filename)
    return fig


def plot_roc(
    truth: Sequence,
    prediction_probabilities: Sequence,
    filename: str | pathlib.Path | None = None,
    font_size: int = 12,
    figure_size: Optional[Sequence] = None,
    axes: plt.Axes | None = None,
    fig: object | None = None,
    label_x: str | None = "False Positive Rate",
    label_y: str | None = "True Positive Rate",
    return_rates: bool = False,
) -> tuple[plt.Axes] | tuple[plt.Axes, Sequence, Sequence]:
    """
    Plots Receiver Operating Characteristic curve

    Can optionally return true positive and false positive rate
    (if `return_rates=True`)
    Parameters:
    ----------
    truth: true labels
    prediction_probabilities: probabilities for labels
    filename: save figure to this file, not saved if `None`
    font_size: font size of all labels of plot
    figure_size: size of figure [width, height], default: [6, 6]
    return_rates: return true positive and false positive rate

    Returns
    -------
    axes of plot
    """
    if figure_size is None:
        figure_size = [6, 6]
    false_positive_rate, true_positive_rate, _ = sklearn.metrics.roc_curve(
        truth, prediction_probabilities, drop_intermediate=False
    )
    roc_auc = sklearn.metrics.auc(false_positive_rate, true_positive_rate)
    fig, axes = kego.plotting.utils_plotting.create_figure_axes(
        figure=fig, axes=axes, figure_size=figure_size, font_size=font_size
    )
    lw = 2
    axes.plot(
        false_positive_rate,
        true_positive_rate,
        lw=lw,
        label="ROC curve (area = %.04f)" % roc_auc,
    )

    axes.plot([0, 1], [0, 1], color="navy", lw=lw, linestyle="--")
    kego.plotting.axes_utils.set_axes(
        axes,
        xlim=[0, 1],
        ylim=[0, 1],
        fontsize=font_size,
        label_x=label_x,
        label_y=label_y,
    )

    axes.legend(loc="lower right")
    kego.plotting.figures.save_figure(fig, filename)
    if return_rates:
        return axes, true_positive_rate, false_positive_rate
    return axes



import itertools

import numpy as np
from datasets import Dataset
from transformers import AutoTokenizer


def get_tokenizer(model):
    return AutoTokenizer.from_pretrained(model)


def tokenize(example, tokenizer, label2id):
    text = []
    # these are at the character level
    labels = []
    for t, l, ws in zip(
        example["tokens"], example["provided_labels"], example["trailing_whitespace"]
    ):
        text.append(t)
        labels.extend([l] * len(t))
        # if there is trailing whitespace
        if ws:
            text.append(" ")
            labels.append("O")
    tokenized = tokenizer("".join(text), return_offsets_mapping=True, truncation=False)
    labels = np.array(labels)
    text = "".join(text)
    token_labels = []
    for start_idx, end_idx in tokenized.offset_mapping:
        # CLS token
        if start_idx + end_idx == 0:
            token_labels.append(label2id["O"])
            continue
        # case when token starts with whitespace
        if text[start_idx].isspace():
            start_idx += 1
        while start_idx >= len(labels):
            start_idx -= 1
        token_labels.append(label2id[labels[start_idx]])
    length = len(tokenized.input_ids)
    return {**tokenized, "labels": token_labels, "length": length}


def get_labeling(json):
    all_labels = sorted(list(set(itertools.chain(*[x["labels"] for x in json]))))
    label2id = {l: i for i, l in enumerate(all_labels)}
    id2label = {v: k for k, v in label2id.items()}
    return all_labels, label2id, id2label


def get_ds(json):
    ds = Dataset.from_dict(
        {
            "full_text": [x["full_text"] for x in json],
            "document": [x["document"] for x in json],
            "tokens": [x["tokens"] for x in json],
            "trailing_whitespace": [x["trailing_whitespace"] for x in json],
            "provided_labels": [x["labels"] for x in json],
        }
    )
    return ds


def check_tokenization(ds, tokenizer, id2label):
    # Confirm that alignment is good
    # run multiple times to see different rows
    x = ds.shuffle()[0]
    for t, l in zip(x["tokens"], x["provided_labels"]):
        if l != "O":
            print((t, l))
    print("*" * 100)
    for t, l in zip(tokenizer.convert_ids_to_tokens(x["input_ids"]), x["labels"]):
        if id2label[l] != "O":
            print((t, id2label[l]))



import json

import arckit
import drawsvg
import numpy as np
from IPython.core.display import SVG, display


def show(drawing: drawsvg.Drawing):
    display(SVG(drawing.as_svg()))


def show_grid(grid):
    for sub_grid in grid:
        print(sub_grid)


def get_solution(task: arckit.data.Task, train_solutions):
    solution = train_solutions[task.id]
    return task, solution


def show_task(task: arckit.data.Task, train_solutions):
    drawing = arckit.vis.draw_task(task, include_test=True)
    solution = train_solutions[task.id]
    show(drawing)
    show_grid(solution[0])
    show(arckit.vis.draw_grid(np.array(solution[0])))
    return drawing



import pandas as pd
from sklearn.model_selection import train_test_split


def split_dataset(
    df: pd.DataFrame,
    train_size: float | None = 0.6,
    validate_size: float | None = 0.2,
    test_size: float | None = None,
    shuffle: bool = True,
    random_state: int = 42,
    stratify_column: str | None = None,
):
    if validate_size is None and test_size is not None and train_size is not None:
        validate_size = 1.0 - test_size - train_size
    elif test_size is None and validate_size is not None and train_size is not None:
        test_size = 1.0 - validate_size - train_size
    elif train_size is None and validate_size is not None and test_size is not None:
        train_size = 1.0 - test_size - validate_size
    if validate_size is None or test_size is None or train_size is None:
        raise ValueError(
            f"Only one of {validate_size=} or {test_size=} or {train_size=} can be None!"
        )
    if validate_size + test_size + train_size != 1:
        raise ValueError(
            f"{validate_size=} + {test_size=} + {train_size=} != {validate_size + test_size + train_size}"
        )

    _train_size, _validate_size, _test_size = (
        round(train_size, 6),
        round(validate_size, 6),
        round(test_size, 6),
    )
    if stratify_column is not None:
        stratify = df[stratify_column]
    train, validate_test = train_test_split(
        df,
        train_size=_train_size,
        test_size=round(1 - _train_size, 6),
        random_state=random_state,
        stratify=stratify,
        shuffle=shuffle,
    )
    if _test_size != 0:
        _validate_size = _validate_size / (_validate_size + _test_size)
        _test_size = 1 - _validate_size
        _validate_size, _test_size = round(_validate_size, 6), round(_test_size, 6)
        if stratify_column is not None:
            stratify = validate_test[stratify_column]
        validate, test = train_test_split(
            validate_test,
            train_size=_validate_size,
            test_size=_test_size,
            random_state=random_state,
            stratify=stratify,
            shuffle=shuffle,
        )
    else:
        validate = validate_test
        test = None
    print(
        f"Split as train: {train.shape[0]} ({train_size*100}%), "
        f"validate: {validate.shape[0]} ({validate_size*100}%), "
        f"test: {test.shape[0] if test is not None else None} ({test_size*100 if test_size is not None else None}%)"
    )
    return train, validate, test


def flatten(l):
    return [item for sublist in l for item in sublist]


def build_xy(
    *dfs,
    y_columns: list[str] | str,
    xy_columns: list[str] | None = None,
    drop_columns: list[str] | None = None,
):
    XX = []
    yy = []
    if isinstance(y_columns, str):
        y_columns = [y_columns]
    if drop_columns is not None and y_columns is not None:
        drop_columns = list(drop_columns) + y_columns
    else:
        drop_columns = y_columns
    for df in dfs:
        df = df.copy()
        y = df[y_columns].values.ravel()
        X = df.drop(columns=drop_columns)
        if xy_columns is not None:
            for column in xy_columns:
                y[column] = X[column]

        XX.append(X)
        yy.append(y)
    return flatten([(X, y) for X, y in zip(XX, yy)])



import kego.plotting.analysis
import kego.plotting.axes_utils
import kego.plotting.colormesh
import kego.plotting.figures
import kego.plotting.histograms
import kego.plotting.lines
import kego.plotting.timeseries
import kego.plotting.value_counts
from kego.plotting.histograms import plot_histogram, plot_histogram_2d



import torch


def available_gpus():
    available_gpus = [
        torch.cuda.get_device_name(i) for i in range(torch.cuda.device_count())
    ]
    return available_gpus


def current_device():
    return torch.cuda.current_device()


def info_gpus():
    print(f"{current_device()=}")
    print(f"{available_gpus()=}")


def empty_gpu():
    torch.cuda.empty_cache()



def assert_missing_columns(df, required_columns: list[str] | None, df_name: str = "df"):
    if required_columns is None:
        return
    missing_columns = set(required_columns) - set(df.columns)
    if missing_columns:
        raise ValueError(
            f"{missing_columns=} in `{df_name}` with {df.columns=} but {required_columns=}"
        )
    return



def train_model(
    model,
    X_train,
    y_train,
    ignore_columns: None | list[str] = None,
    feature_selector=None,
    feature_space=None,
):
    X_train = X_train.copy()
    y_train = y_train.copy()
    if ignore_columns is not None:
        X_train = X_train.drop(columns=ignore_columns)
    if feature_selector is not None:
        if feature_space is None:
            raise ValueError(f"Define `feature_space` when using {feature_selector=}.")
        model = feature_selector(model, feature_space)
    model = model.fit(X_train, y_train)
    return model



from typing import Iterable, overload

import numpy as np


def flatten_list(lis: list[list]) -> list:
    return [item for sublist in lis for item in sublist]


def to_nlength_tuple(x: object, n: int = 2) -> tuple:
    if not (isinstance(x, list) or isinstance(x, tuple) or isinstance(x, np.ndarray)):
        return tuple([x] * n)
    elif isinstance(x, list) or isinstance(x, np.ndarray):
        x = tuple(x)
    if len(x) != n:
        raise ValueError(f"{x} doesn't have expected length {n=}")
    return x


def is_in_list_and_remove(to_check, _list):
    is_in = False
    if to_check in _list:
        is_in = True
    if is_in:
        _list.remove(to_check)
    return is_in


def is_listlike(to_check: object):
    if isinstance(to_check, Iterable) and not isinstance(to_check, str):
        return True
    return False



def break_lines(string, max_length=100):
    all_lines = ""
    current_line = ""
    for word in string.split(" "):
        if len(current_line + " " + word) < max_length:
            current_line += " " + word
        else:
            all_lines += "\n" + current_line
            current_line = word
    all_lines += "\n" + current_line
    all_lines = all_lines.strip("\n ")
    return all_lines



import itertools
import logging
from typing import Sequence

import numpy as np

import kego.plotting.utils_plotting
from kego.lists import is_listlike, to_nlength_tuple
from kego.plotting.figures import plot_legend, save_figure
from kego.plotting.utils_plotting import create_axes_grid


def plot_lines(
    xs: np.ndarray,
    ys: np.ndarray,
    labels: None | tuple[np.ndarray] = None,
    nx_max=4,
    filename: str | None = None,
    log: str | tuple[str, str] = "false",
):
    assert len(np.shape(xs)) != 2, f"xs needs to be of shape 2 but is {xs.shape=}!"
    assert len(np.shape(ys)) != 2, f"xs needs to be of shape 2 but is {ys.shape=}!"

    n_plots = np.shape(xs)[0]
    labels = to_nlength_tuple(labels, n_plots)
    if n_plots >= nx_max:
        n_rows = n_plots
    n_columns = np.ceil(n_plots / nx_max)
    figure, axes_grid, _ = create_axes_grid(n_columns=n_columns, n_rows=n_rows)
    for i_plot, (i_row, i_column) in enumerate(
        itertools.product(range(n_rows), range(n_columns))
    ):
        axes = axes_grid[i_row, i_column]
        plot_line(x=xs[i_plot], y=ys[i_plot], label=labels[i_plot], axes=axes)
        plot_legend(axes=axes)
    save_figure(fig=figure, filename=filename)


def plot_line(
    x: np.ndarray,
    y: np.ndarray,
    xlim: tuple[float | None, float | None] | None = None,
    ylim: tuple[float | None, float | None] | None = None,
    log: str | tuple[str, str] = "false",
    label: str | None = None,
    replace_x_labels_at: Sequence | None = None,
    replace_x_labels_with: Sequence | None = None,
    replace_y_labels_at: Sequence | None = None,
    replace_y_labels_with: Sequence | None = None,
    rotation_x_labels: int = 0,
    rotation_y_labels: int = 0,
    filename: str | None = None,
    axes=None,
    font_size: int = 12,
    symlog_linear_threshold: float | None = None,
    label_x: str | None = None,
    label_y: str | None = None,
):
    if axes is None:
        figure, axes, _ = create_axes_grid(n_columns=1, n_rows=1, unravel=True)
    _log = to_nlength_tuple(log)
    _xlim = to_nlength_tuple(xlim)
    _ylim = to_nlength_tuple(ylim)
    axes.plot(x, y, label=label)
    kego.plotting.utils_plotting.set_x_log(
        axes, _log[0], axis_symlog_linear_threshold=symlog_linear_threshold
    )
    kego.plotting.utils_plotting.set_y_log(
        axes, _log[1], axis_symlog_linear_threshold=symlog_linear_threshold
    )
    kego.plotting.utils_plotting.set_axis_tick_labels(
        axes,
        replace_x_labels_at,
        replace_x_labels_with,
        axis="x",
        rotation=rotation_x_labels,
        font_size=font_size,
    )
    kego.plotting.utils_plotting.set_axis_tick_labels(
        axes,
        replace_y_labels_at,
        replace_y_labels_with,
        axis="y",
        rotation=rotation_y_labels,
        font_size=font_size,
    )
    kego.plotting.utils_plotting.set_axes_label(axes, label_x, "x", font_size=font_size)
    kego.plotting.utils_plotting.set_axes_label(axes, label_y, "y", font_size=font_size)
    axes.set_ylim(_ylim)
    axes.set_xlim(_xlim)
    save_figure(fig=figure, filename=filename)
    return axes



import numpy as np
import pandas as pd

import kego.plotting


def plot_value_counts(
    df: pd.DataFrame,
    rotation_x_labels: int = 90,
    rotation_y_labels: int = 0,
    font_size: int = 12,
    figure_size: tuple | None = None,
):
    for column in df.columns:
        figure, axes, _ = kego.plotting.utils_plotting.create_axes_grid(
            n_columns=1, n_rows=1, unravel=True, figure_size=figure_size
        )
        value_counts = df[column].value_counts()
        values = value_counts.index
        counts = value_counts.values
        bin_edges = np.arange(len(values) + 1)
        width_bars = np.diff(bin_edges)
        bin_centers = bin_edges[:-1] + width_bars

        replace_x_labels_at = None
        replace_x_labels_with = None
        if len(values) < 100:
            replace_x_labels_at = bin_centers
            replace_x_labels_with = values

            kego.plotting.histograms.plot_bar(
                bin_centers=bin_centers,
                height_bars=counts,
                width_bars=width_bars,
                replace_x_labels_at=replace_x_labels_at,
                replace_x_labels_with=replace_x_labels_with,
                rotation_x_labels=rotation_x_labels,
                rotation_y_labels=rotation_y_labels,
                axes=axes,
                font_size=font_size,
            )
        else:
            kego.plotting.histograms.plot_histogram(
                counts,
                axes=axes,
                font_size=font_size,
            )

    return figure



import logging
from functools import wraps
from time import time
from typing import Callable


def timing(f: Callable):
    """
    Wrapper that returns execution time and arguments of function.
    """

    @wraps(f)
    def wrap(*args, **kw):
        ts = time()
        result = f(*args, **kw)
        te = time()
        logging.info(f"func:{f.__name__} args:[{kw}] took: {te-ts} sec")
        return result

    return wrap



from typing import Any

import pandas as pd

from kego.columns import assert_missing_columns
from kego.dataframes import retain_df


def from_prediction_to_solution_submission(
    y_true: pd.DataFrame,
    y_prediction: pd.DataFrame | None = None,
    X_prediction: pd.DataFrame | None = None,
    row_id_column_name: str = "id",
    model=None,
    prediction_column: str = "prediction",
    target_column: str | None = None,
    required_columns: list[str] | None = None,
    dummy_column_name_value: dict[str, Any] | None = None,
):
    if y_prediction is None and X_prediction is None:
        raise ValueError(f"Need to provide {X_prediction} or {y_prediction}")
    if y_prediction is None and X_prediction is not None:
        if model is None or target_column is None:
            raise ValueError(
                f"Need to provide `model` and `target_column` when `X_prediction` given."
            )
        else:
            prediction_values = model.predict(X_prediction.copy())
            y_prediction = pd.DataFrame(data={target_column: prediction_values})
    if not isinstance(y_true, pd.DataFrame):
        y_true = pd.DataFrame(data={target_column: y_true})
    solution, submission, X_prediction = (
        retain_df(y_true),
        retain_df(y_prediction),
        retain_df(X_prediction),
    )
    if required_columns is not None:
        for column in required_columns:
            solution[column] = X_prediction[column]
    if dummy_column_name_value is not None:
        for dummy_column_key, dummy_column_value in dummy_column_name_value.items():
            solution[dummy_column_key], submission[dummy_column_key] = (
                dummy_column_value,
                dummy_column_value,
            )
    submission = submission.rename(columns={target_column: prediction_column})
    assert len(solution) == len(submission)
    submission.insert(0, row_id_column_name, range(len(submission)))
    solution.insert(0, row_id_column_name, range(len(solution)))
    return solution, submission



import itertools
import multiprocessing
from typing import Callable, Iterable, Optional


def get_processes(processes):
    if processes < 0:
        processes = multiprocessing.cpu_count() + processes + 1
    return processes


def parallelize(
    function: Callable,
    args_zipped: Iterable,
    processes: int = -1,
    single_arg: bool = False,
    kwargs_as_dict: Optional[dict] = None,
):
    """
    parallelize function with args provided in zipped format
    ----------
    function: function to parallelize
    args: args of function in zipped format

    Returns
    -------
    function applied to args
    """
    if processes == 1:
        results = []
        if kwargs_as_dict is None:
            kwargs_as_dict = {}
        for args in args_zipped:
            if isinstance(args, str):
                args = (args,)
            results.append(function(*args, **kwargs_as_dict))
        return results
    processes = get_processes(processes)
    print(f"n processes: {processes}")
    if single_arg and kwargs_as_dict is None:
        with multiprocessing.Pool(processes=processes) as pool:
            results = pool.map(function, args_zipped)
    else:
        if single_arg:
            args_zipped = ((arg,) for arg in args_zipped)
        if kwargs_as_dict is not None:
            kwargs_iter = itertools.repeat(kwargs_as_dict)
        else:
            kwargs_iter = itertools.repeat(dict())
        with multiprocessing.Pool(processes=processes) as pool:
            results = starmap_with_kwargs(pool, function, args_zipped, kwargs_iter)
    return results


def starmap_with_kwargs(
    pool, function: Callable, args_iter: Iterable, kwargs_iter: Iterable
):
    """Helper function to parallelize functions with args and kwargs"""
    if kwargs_iter is None:
        args_for_starmap = zip(itertools.repeat(function), args_iter)
    else:
        args_for_starmap = zip(itertools.repeat(function), args_iter, kwargs_iter)
    return pool.starmap(apply_args_and_kwargs, args_for_starmap)


def apply_args_and_kwargs(fn: Callable, args, kwargs):
    """Helper function to parallelize functions with args and kwargs"""
    return fn(*args, **kwargs)



import multiprocessing
import os

import dotenv
import pandas as pd
import tqdm


def get_phase(i, planet_id, n_steps, dataset, adc_info):
    f_signal = pd.read_parquet(
        os.environ["FOLDER_COMPETITION"] + f"{dataset}/{planet_id}/FGS1_signal.parquet"
    )
    mean_signal = f_signal.values.mean(axis=1)  # mean over the 32*32 pixels
    net_signal = mean_signal[1::2] - mean_signal[0::2]
    gain = adc_info.FGS1_adc_gain.values[i]
    return [
        net_signal[i * n_steps : (i + 1) * n_steps].mean() * gain
        for i in range(len(net_signal) // n_steps + 1)
    ]


if __name__ == "__main__":
    dotenv.load_dotenv(override=True, verbose=True)
    dataset = "train"
    train_adc_info = pd.read_csv(
        os.environ["FOLDER_COMPETITION"] + "train_adc_info.csv", index_col="planet_id"
    )
    adc_info = train_adc_info
    for n_steps in [100, 400, 1000, 2000, 8000][::-1]:
        print(f"Starting {n_steps=} ...")
        planet_ids = adc_info.index

        indices = range(len(planet_ids))
        print(planet_ids)
        # with multiprocessing.Pool(processes=multiprocessing.cpu_count() - 2) as pool:
        #     result = pool.apply_async(
        phases = []
        for i, planet_id in tqdm.tqdm(zip(indices, planet_ids), total=len(indices)):
            phase = get_phase(i, planet_id, n_steps, dataset, adc_info)
        phases.append(phase)
        # phases = result.get()
        df = pd.DataFrame(
            phases, columns=[f"phase_{i}" for i in range(len(phases[0]))], index=indices
        )
        filepath = f"phases_step{n_steps}_nids{len(indices)}.csv"
        print(f"... saving {filepath}")
        df.to_csv(filepath)



import logging
import pathlib
from collections.abc import Sequence
from typing import Literal

import matplotlib.container
import numpy as np
import pandas as pd

import kego.checks
import kego.constants
import kego.lists
import kego.plotting.axes_utils
import kego.plotting.colormesh
import kego.plotting.figures
import kego.plotting.histogram_2d_utils
import kego.plotting.utils_plotting

logger = logging.getLogger(__name__)


def plot_histogram_2d(
    x_key_or_values: np.ndarray | str,
    y_key_or_values: np.ndarray | str,
    df: pd.DataFrame | None = None,
    bin_edges: tuple[np.ndarray, np.ndarray] | None = None,
    xlim: tuple[float | None, float | None] | None = None,
    ylim: tuple[float | None, float | None] | None = None,
    log: str | tuple[str, str] = "false",
    filename: str | pathlib.Path | None = None,
    n_bins: int | tuple[int, int] = 60,
    n_bins_linear: int = 10,
    norm: str | None = None,
    norm_symlog_linear_threshold: float | None = None,
    axis_symlog_linear_threshold: float | tuple[float, float] = 1e-9,
    label_x: str | None = None,
    label_y: str | None = None,
    label_colorbar: str | None = "counts",
    font_size: float = kego.constants.DEFAULT_FONTSIZE_SMALL,
    annotate: bool = False,
    annotate_round_to_base: int | None = None,
    annotate_color: str = "gray",
    annotate_line: Literal["unity"] | None = None,
    vmin: float | None = None,
    vmax: float | None = None,
    facet_column: str | None = None,
    facet_row: str | None = None,
    facet_show_label_x: list[int] | str = "minimal",
    facet_show_label_y: list[int] | str = "minimal",
    marginal_x: str | None = "default",
    marginal_x_label_x: str | None = None,
    marginal_x_label_y: str | None = "counts",
    marginal_x_show_xticks: bool = False,
    marginal_y: str | None = "default",
    marginal_y_label_x: str | None = "counts",
    marginal_y_label_y: str | None = None,
    marginal_y_show_yticks: bool = False,
    marginal_color: str | None = "#1f77b4",
    figure_size: tuple[float, float] | None = None,
    colormap: str = kego.constants.DEFAULT_COLORMAP,
    colorbar_width: float = 0.02,
    spacing_x: float | None = None,
    spacing_y: float | None = None,
    spacing_colorbar: float = 0.03,
    title: str = "",
    convert_zeros_to_nan: bool = True,
    axes: kego.constants.TYPE_MATPLOTLIB_AXES | None = None,
    figure: kego.constants.TYPE_MATPLOTLIB_FIGURES | None = None,
    axes_colorbar: kego.constants.TYPE_MATPLOTLIB_AXES | None = None,
) -> dict:
    """
    Plots 2d histogram including histograms in margins and facet

    Log (axis) types included "false", "symlog", "log"
    Norm takes care of colorbar scale ("log", "linear"/None, "symlog")
    Allows for marginal 1D histograms on top and to left,
        set `marginal_x="histogram"` and `marginal_y="histogram"`
    Can annotate values on 2d histogram, set `annotate="True"`
    Allows splitting of 2d histogram into unique values of variable
        along rows (facet_row) and/or columns (facet_column)
        (Cannot be used with `marginal_x`/`marginal_y`)

    Parameters:
    ----------
    x_key_or_values:
        Values binned along x-axis, array(N) or column name found in `df`
    y_key_or_values:
        Values binned along y-axis, array(N) or column name found in `df`
    df:
        Dataframe used for `x`, `y`, `facet_column`, `facet_row`
    bin_edges:
        Bin edges provided as [x_edges, y_edges],
        if `None` computed based on data and `xlim`/`ylim` provided and `n_bins`.
    xlim:
        Limits of x-axis and bin_edges, if None determined from (min, max) of x-values
    ylim:
        Limits of y-axis and bin_edges, if None determined from (min, max) of y-values
    log:
        Type of bins ("false", "symlog", "log")
    filename:
        Figure saved to file if provided
    n_bins:
        Number of bins if `bin_edges` is None,
        can be provided as (x-axis n_bins, y-axis n_bins) or n_bins used for both x-axis/y-axis
    n_bins_linear:
        Only relevant if `log="symlog"`, Number of bins for linear part of bins (around zero)
    norm:
        Scaling of colorbar (None/"linear", "log", "symlog")
    norm_symlog_linear_threshold:
        Required when `norm="symlog"`.
        Threshold value below which symlog of norm becomes linear.
    axis_symlog_linear_threshold:
        Required when `log="symlog"`.
        Threshold value below which symlog of axis becomes linear.
    label_x:
        Label of x-axis
    label_y:
        Label of y-axis
    label_colorbar:
        Label of colorbar
    font_size:
        Size of font
    annotate:
        Whether to show number of samples on plot
    annotate_round_to_base:
        Round annotated values to this base integer
    annotate_color:
        Color of samples to be annotated
    vmin:
        Minimum value of colormap
    vmax:
        Maximum value of colormap
    facet_column:
        Assign all samples belonging to unique value from this field to histogram as columns
    facet_row:
        Assign all samples belonging to unique value from this field to histogram as rows
    facet_show_label_x:
        Which x-labels to show along rows provided as list of integers or use presets "off"/"all"/"minimal".
    facet_show_label_y:
        Which y-labels to show along columns provided as list of integers or use presets "off"/"all"/"minimal".
    marginal_x:
        Add additional plot next to histogram plot ("histogram"/"default"/None)
    marginal_x_label_x:
        X-axis label for plot `marginal_x`
    marginal_x_label_y:
        Y-axis label for plot `marginal_x`
    marginal_x_show_xticks:
        Wether to show ticks along the x-axis for plot `marginal_x`
    marginal_y:
        Add additional plot on top of histogram plot ("histogram"/"default"/None)
    marginal_y_label_x:
        X-axis label for plot `marginal_y`
    marginal_y_label_y:
        Y-axis label for plot `marginal_y`
    marginal_y_show_yticks:
        Wether to show ticks along the y-axis for plot `marginal_y`
    marginal_color:
        Color of histogram bars in marginal plots
    figure_size:
        Size of figure (height, width)
    colormap:
        Matplotlib colormap name
    colorbar_width:
        Width of the colorbars
    spacing_x:
        Spacing between axes along horizontal direction
    spacing_y:
        Spacing between axes along vertical direction
    spacing_colorbar:
        Spacing between axes and colorbar axes
    title:
        Title of figure
    convert_zeros_to_nan:
        Whether to convert all zero values in histogram 2d to nan.
        This will exclude these values from being colored according to colormap but render them white.
    axes:
        Plots 2d histogram on specified axes.
        Excludes usage of `marginal_x`, `marginal_y`, `facet_column` and/or `facet_row`.
    figure:
        Used as matplotlib figure for plot of 2d histogram when `axes` specified.
    axes_colorbar:
        Used as axes for colorbar of 2d histogram when `axes` specified.

    Returns
    -------
    Dictionary of axes `axes_2d_histograms_{i_row}{i_column}`
    and histograms `2d_histograms_{i_row}{i_column}`
    """

    axes_style = kego.plotting.histogram_2d_utils._resolve_axes_style(axes)

    (
        no_marginal_plots,
        marginal_x,
        marginal_y,
        spacing_x,
        spacing_y,
    ) = kego.plotting.histogram_2d_utils._resolve_defaults(
        facet_column,
        facet_row,
        marginal_x,
        marginal_y,
        spacing_x,
        spacing_y,
        axes_style,
    )
    kego.plotting.histogram_2d_utils._check_parameter_consistency(
        facet_column, facet_row, marginal_x, marginal_y, df
    )

    (
        column_masks,
        n_facet_column_values,
        facet_column_title,
    ) = kego.plotting.histogram_2d_utils._prepare_facet(df, facet_column)
    (
        row_masks,
        n_facet_row_values,
        facet_row_title,
    ) = kego.plotting.histogram_2d_utils._prepare_facet(df, facet_row)

    (
        n_columns,
        n_rows,
    ) = kego.plotting.histogram_2d_utils._set_number_rows_and_columns(
        marginal_x, marginal_y, n_facet_column_values, n_facet_row_values
    )

    (
        xaxis_labels,
        yaxis_labels,
    ) = kego.plotting.histogram_2d_utils._set_axis_labels(
        n_columns,
        n_rows,
        facet_show_label_x,
        facet_show_label_y,
        label_x,
        label_y,
        facet_column,
        facet_row,
    )

    masks = kego.plotting.histogram_2d_utils._set_masks(
        column_masks, row_masks, n_columns, n_rows
    )

    titles = kego.plotting.histogram_2d_utils._determine_axes_titles(
        facet_column,
        facet_row,
        title,
        facet_column_title,
        facet_row_title,
        n_columns,
        n_rows,
    )

    (
        skip_row_col,
        widths_along_x,
        heights_along_y,
        colorbar_include_row_col,
        colorbar_off,
    ) = kego.plotting.histogram_2d_utils._prepare_axes_grid_creation(
        marginal_x, marginal_y
    )
    if axes_style == "axes_grid":
        (
            fig,
            axes_grid,
            axes_colorbar_grid,
        ) = kego.plotting.utils_plotting.create_axes_grid(
            n_columns=n_columns,
            n_rows=n_rows,
            figure_size=figure_size,
            skip_row_column=skip_row_col,
            colorbar_width=colorbar_width,
            spacing_x=spacing_x,
            spacing_y=spacing_y,
            widths_along_x=widths_along_x,
            heights_along_y=heights_along_y,
            colorbar_include_row_columns=colorbar_include_row_col,
            spacing_colorbar=spacing_colorbar,
            colorbar_off=colorbar_off,
            left=0.08,
            bottom=0.1,
        )
    else:
        fig, axes = kego.plotting.utils_plotting.create_figure_axes(
            figure=figure, axes=axes
        )
    results_dic = {}
    for i_row in range(n_rows):
        for i_column in range(n_columns):
            axes_marginal = None
            if axes_style == "axes_grid":
                axes = axes_grid[i_row, i_column]
                axes_colorbar = axes_colorbar_grid[i_row, i_column]
                if not no_marginal_plots:
                    axes = axes_grid[1][0]
                    axes_colorbar = axes_colorbar_grid[1][1]
                    axes_marginal = axes_grid
            mask = masks[i_row, i_column]
            title = titles[i_row, i_column]
            _ax, _hist = _plot_histogram_2d(
                x_key_or_values,
                y_key_or_values,
                axes=axes,
                axes_colorbar=axes_colorbar,
                df=df,
                bin_edges=bin_edges,
                xlim=xlim,
                ylim=ylim,
                log=log,
                n_bins=n_bins,
                n_bins_linear=n_bins_linear,
                norm=norm,
                norm_symlog_linear_threshold=norm_symlog_linear_threshold,
                linear_thresh=axis_symlog_linear_threshold,
                label_x=xaxis_labels[i_row, i_column],
                label_y=yaxis_labels[i_row, i_column],
                label_colorbar=label_colorbar,
                font_size=font_size,
                annotate=annotate,
                annotate_round_to_base=annotate_round_to_base,
                annotate_color=annotate_color,
                vmin=vmin,
                vmax=vmax,
                marginal_x=marginal_x,
                marginal_x_label_x=marginal_x_label_x,
                marginal_x_label_y=marginal_x_label_y,
                marginal_x_show_xticks=marginal_x_show_xticks,
                marginal_y=marginal_y,
                marginal_y_label_x=marginal_y_label_x,
                marginal_y_label_y=marginal_y_label_y,
                marginal_y_show_yticks=marginal_y_show_yticks,
                marginal_color=marginal_color,
                axes_marginal=axes_marginal,
                colormap=colormap,
                title=title,
                mask=mask,
                convert_zeros_to_nan=convert_zeros_to_nan,
            )
            if no_marginal_plots:
                results_dic[f"axes_2d_histograms_{i_row}{i_column}"] = _ax
                results_dic[f"2d_histograms_{i_row}{i_column}"] = _hist
            else:
                results_dic["axes_2d_histograms_10"] = _ax
                results_dic["2d_histograms_10"] = _hist
                results_dic["axes_histograms_00"] = axes_grid[0, 0]
                results_dic["axes_histograms_11"] = axes_grid[1, 1]
            if not no_marginal_plots:
                break
            add_annotation_line(style=annotate_line, axes=axes)
        if not no_marginal_plots:
            break
    kego.plotting.figures.save_figure(fig, filename)
    return results_dic


def add_annotation_line(
    style: Literal["unity"] | None, axes, alpha: float = 0.4, plot_style: str = "--k"
):
    if style is None:
        return
    if style == "unity":
        xlim = axes.get_xlim()
        ylim = axes.get_ylim()
        axes.plot(xlim, ylim, plot_style, label="unity", alpha=alpha)


def _get_label(label: str | None) -> str:
    if label is not None and label:
        return label
    else:
        return ""


def _get_values_and_axis_label_from_dataframe(
    df: pd.DataFrame | None, key_or_values: np.ndarray | str, label: str | None
) -> tuple[np.ndarray, str | None]:
    if isinstance(key_or_values, str):
        if df is not None:
            if label is None:
                label = key_or_values
            if label == "off":
                label = None
            values = df[key_or_values].values
        else:
            raise ValueError(
                f"{df=}, if {key_or_values=} given as string, dataset required"
            )
    else:
        values = key_or_values
    return values, label


def _plot_histogram_2d(
    x_key_or_values: np.ndarray | str,
    y_key_or_values: np.ndarray | str,
    axes: kego.constants.TYPE_MATPLOTLIB_AXES,
    axes_colorbar: kego.constants.TYPE_MATPLOTLIB_AXES,
    xlim: tuple[float | None, float | None] | None = None,
    ylim: tuple[float | None, float | None] | None = None,
    df: pd.DataFrame | None = None,
    bin_edges: tuple[np.ndarray, np.ndarray] | None = None,
    log: str | tuple[str, str] = "false",
    fig: kego.constants.TYPE_MATPLOTLIB_FIGURES | None = None,
    n_bins: int | tuple[int, int] = 60,
    n_bins_linear: int = 10,
    norm: str | None = None,
    norm_symlog_linear_threshold: float | None = None,
    linear_thresh: float | tuple[float, float] = 1e-9,
    label_x: str | None = None,
    label_y: str | None = None,
    label_colorbar: str | None = None,
    font_size: float = kego.constants.DEFAULT_FONTSIZE_SMALL,
    annotate: bool = False,
    annotate_round_to_base: int | None = None,
    annotate_color: str = "gray",
    vmin: float | None = None,
    vmax: float | None = None,
    marginal_x: str | None = "histogram",
    marginal_x_label_x: str | None = None,
    marginal_x_label_y: str | None = "N",
    marginal_x_show_xticks: bool = False,
    marginal_y: str | None = "histogram",
    marginal_y_label_x: str | None = "N",
    marginal_y_label_y: str | None = None,
    marginal_y_show_yticks: bool = False,
    marginal_color: str | None = None,
    colormap: str = kego.constants.DEFAULT_COLORMAP,
    title: None | str = None,
    axes_marginal: list | None = None,
    mask: np.ndarray | None = None,
    convert_zeros_to_nan: bool = True,
) -> tuple[kego.constants.TYPE_MATPLOTLIB_AXES, np.ndarray]:
    """
    Plots 2d histogram

    Parameters:
    ----------
    axes_marginal:
        Matplotlib axes to plot margins into
    mask:
        Mask x/y values (require same shape as x/y)

    See definition of `plot_histogram_2d` for explanation of parameters

    Returns
    -------
    Axes, histogram values
    """

    x_values, label_x = _get_values_and_axis_label_from_dataframe(
        df, x_key_or_values, label_x
    )
    y_values, label_y = _get_values_and_axis_label_from_dataframe(
        df, y_key_or_values, label_y
    )

    x_values, y_values = _get_xy_values(x_values, y_values, mask)

    xlim, ylim, log, n_bins, n_bins_linear, linear_thresh = _prepare_parameters(  # type: ignore
        xlim, ylim, log, n_bins, n_bins_linear, linear_thresh
    )

    if bin_edges is None:
        bin_edges_x = get_bin_edges(
            data=x_values,
            n_bins=n_bins[0],
            n_bins_linear=n_bins_linear[0],  # type: ignore
            symlog_linear_threshold=linear_thresh[0],
            log=log[0],
            vmin=xlim[0],
            vmax=xlim[1],
        )
        bin_edges_y = get_bin_edges(
            data=y_values,
            n_bins=n_bins[1],
            n_bins_linear=n_bins_linear[1],  # type: ignore
            symlog_linear_threshold=linear_thresh[1],
            log=log[1],
            vmin=ylim[0],
            vmax=ylim[1],
        )
    else:
        bin_edges_x, bin_edges_y = bin_edges

    kego.checks.validate_array(bin_edges_x, name="bin_edges_x")  # type: ignore
    kego.checks.validate_array(bin_edges_y, name="bin_edges_y")  # type: ignore

    norm_object = kego.plotting.utils_plotting.get_norm(
        norm,
        vmin=vmin,
        vmax=vmax,
        norm_symlog_linear_threshold=norm_symlog_linear_threshold,
    )
    H, bin_edges_x, bin_edges_y = np.histogram2d(
        x_values,
        y_values,
        bins=[np.array(bin_edges_x), np.array(bin_edges_y)],
    )
    H_plot = H.T
    X, Y = np.meshgrid(bin_edges_x, bin_edges_y)
    plot = kego.plotting.colormesh._plot_colormesh(
        axes,
        X,
        Y,
        H_plot,
        norm_object,
        colormap,
        linewidth=0,
        rasterized=True,
        convert_zeros_to_nan=convert_zeros_to_nan,
    )
    if annotate:
        kego.plotting.utils_plotting.annotate_values(
            H=H_plot,
            axes=axes,
            size_x=len(bin_edges_x) - 1,
            size_y=len(bin_edges_y) - 1,
            color=annotate_color,
            round_to_base=annotate_round_to_base,
            font_size=font_size,
        )
    kego.plotting.utils_plotting.plot_colorbar(
        plot, cax=axes_colorbar, label=label_colorbar, font_size=font_size
    )

    xlim, ylim = _find_axis_limits(xlim, ylim, bin_edges_x, bin_edges_y)

    kego.plotting.utils_plotting.set_x_log(
        axes, log[0], axis_symlog_linear_threshold=linear_thresh[0]
    )
    kego.plotting.utils_plotting.set_y_log(
        axes, log[1], axis_symlog_linear_threshold=linear_thresh[1]
    )
    kego.plotting.axes_utils.set_x_lim(axes, xlim)
    kego.plotting.axes_utils.set_y_lim(axes, ylim)

    kego.plotting.utils_plotting.set_title(axes, title, font_size=font_size)

    if marginal_x == "histogram" and axes_marginal is not None:
        plot_histogram(
            x_values,
            bin_edges=bin_edges_x,
            xlim=xlim,
            ylim=None,
            axes=axes_marginal[0][0],
            log=log[0],
            figure=fig,
            n_bins=n_bins[0],
            n_bins_linear=n_bins_linear[0],  # type: ignore
            symlog_linear_threshold=linear_thresh[0],
            label_x=_get_label(marginal_x_label_x),
            label_y=_get_label(marginal_x_label_y),
            font_size=font_size,
            color=marginal_color,
            tight_layout=False,
        )
        if not marginal_x_show_xticks:
            kego.plotting.utils_plotting.remove_tick_labels(axes_marginal[0][0], "x")
    if marginal_y == "histogram" and axes_marginal is not None:
        plot_histogram(
            y_values,
            bin_edges=bin_edges_y,
            xlim=None,
            ylim=ylim,
            axes=axes_marginal[1][1],
            log=log[1],
            figure=fig,
            n_bins=n_bins[1],
            n_bins_linear=n_bins_linear[1],  # type: ignore
            symlog_linear_threshold=linear_thresh[1],
            label_x=_get_label(marginal_y_label_x),
            label_y=_get_label(marginal_y_label_y),
            font_size=font_size,
            vertical=True,
            color=marginal_color,
            tight_layout=False,
            rotation_x_labels=270,
        )
        if not marginal_y_show_yticks:
            kego.plotting.utils_plotting.remove_tick_labels(axes_marginal[1][1], "y")
    kego.plotting.utils_plotting.set_axis_tick_labels(
        axes, font_size=font_size, axis="x"
    )
    kego.plotting.utils_plotting.set_axis_tick_labels(
        axes, font_size=font_size, axis="y"
    )
    kego.plotting.utils_plotting.set_axes_label(axes, label_x, "x", font_size=font_size)
    kego.plotting.utils_plotting.set_axes_label(axes, label_y, "y", font_size=font_size)
    return axes, H


def _prepare_parameters(
    xlim: tuple[float | None, float | None] | None,
    ylim: tuple[float | None, float | None] | None,
    log: str | tuple[str, str],
    n_bins: int | tuple[int, int],
    n_bins_linear: int | tuple[int, int],
    linear_thresh: float | tuple[float, float],
) -> tuple[
    tuple[float | None, float | None],
    tuple[float | None, float | None],
    tuple[str, str],
    tuple[int, int],
    tuple[int, int],
    tuple[float, float],
]:
    log = kego.lists.to_nlength_tuple(log)  # type: ignore
    n_bins = kego.lists.to_nlength_tuple(n_bins)  # type: ignore
    n_bins_linear = kego.lists.to_nlength_tuple(n_bins_linear)  # type: ignore
    xlim = kego.lists.to_nlength_tuple(xlim)  # type: ignore
    ylim = kego.lists.to_nlength_tuple(ylim)  # type: ignore
    n_bins = tuple(
        [i + 1 if i is not None else i for i in n_bins]  # type: ignore
    )  # using bin edges later, where n_edges = n_bins + 1
    n_bins_linear = tuple(
        [i + 1 if i is not None else i for i in n_bins_linear]  # type: ignore
    )  # using bin edges later, where n_edges = n_bins + 1
    linear_thresh = kego.lists.to_nlength_tuple(linear_thresh)  # type: ignore
    return xlim, ylim, log, n_bins, n_bins_linear, linear_thresh


def _get_xy_values(
    x: np.ndarray, y: np.ndarray, mask: np.ndarray | None
) -> tuple[np.ndarray, np.ndarray]:
    x = np.ndarray.flatten(x)  # type: ignore
    kego.checks.validate_array(x)
    y = np.ndarray.flatten(y)  # type: ignore
    kego.checks.validate_array(y)
    if x.shape != y.shape:
        raise ValueError(
            f"x and y need to be of same shape: {np.shape(x)} != {np.shape(y)}"
        )
    if mask is not None:
        x = x[mask]
        y = y[mask]
    return x, y


def plot_histogram(
    key_or_values: np.ndarray | str,
    df: pd.DataFrame | None = None,
    xlim: tuple[float | None, float | None] | None = None,
    ylim: tuple[float | None, float | None] | None = None,
    log: str | tuple[str, str] = "false",
    symlog_linear_threshold: float | None = None,
    bin_edges: np.ndarray | None = None,
    n_bins: int = 60,
    n_bins_linear: int = 10,
    label_x: str | None = None,
    label_y: str | None = "counts",
    figure: kego.constants.TYPE_MATPLOTLIB_FIGURES | None = None,
    axes: kego.constants.TYPE_MATPLOTLIB_AXES | None = None,
    filename: str | pathlib.Path | None = None,
    font_size: float = kego.constants.DEFAULT_FONTSIZE_LARGE,
    vertical: bool = False,
    alpha: float = 1,
    color: str | None = None,
    horizontal_line: int | None = None,
    tight_layout: bool = True,
    rotation_x_labels: int = 0,
    rotation_y_labels: int = 0,
    title: str = "",
) -> kego.constants.TYPE_MATPLOTLIB_AXES:
    """
    plots 1d histogram

    Log (axis) types included "false", "symlog", "log"
    Norm takes care of colorbar scale ("log", "linear"/None, "symlog")
    Can include a horizontal line at value `horizontal_line`

    Parameters:
    ----------
    x_key_or_values:
        Values binned along x-axis, array(N) or column name found in `df`
    df:
        Dataframe used for `key_or_values`
    xlim:
        Limits of x-axis and bin_edges, if None determined from (min, max) of x-values
    ylim:
        Limits of y-axis and bin_edges, if None determined from (min, max) of y-values
    log:
        Type of bins ("false", "symlog", "log")
    symlog_linear_threshold:
        Required when `log="symlog"`.
        Threshold value below which symlog becomes linear.
    bin_edges:
        Bin edges, if `None` computed based on data `xlim` provided and `n_bins`.
    n_bins:
        Number of bins if `bin_edges` is None.
    n_bins_linear:
        Only relevant if `log="symlog"`, Number of bins for linear part of bins (around zero)
    label_x:
        Label of x-axis
    label_y:
        Label of y-axis
    figure:
        Matplotlib figure
    axes:
        Matplotlib axes
    filename:
        Figure saved to file if provided
    font_size:
        Size of font
    vertical:
        Plot bars along vertical axis instead of default horizontal direction
    alpha:
        Alpha value of bars
    color:
        Color of bars
    horizontal_line:
        Plot horizontal at specified value (not shown if None)
    tight_layout:
        Adjusts padding around axis and removes unnecessary white space
    rotation_x_labels:
        Angle of rotation of x-axis labels
    rotation_y_labels:
        Angle of rotation of y-axis labels
    title:
        Title of figure

    Returns
    -------
    Axes of bar plot
    """
    _log = kego.lists.to_nlength_tuple(log, 2)
    _xlim = kego.lists.to_nlength_tuple(xlim, 2)
    _ylim = kego.lists.to_nlength_tuple(ylim, 2)

    values: np.ndarray
    if df is not None and isinstance(key_or_values, str):
        if label_x is None:
            label_x = key_or_values
        values = df[key_or_values].values
    elif isinstance(key_or_values, str):
        raise ValueError(
            f"Need to specify {df=} when specifying key name {key_or_values=}."
        )
    else:
        values = np.array(key_or_values)
    values = flatten_array(values)
    figure, axes = kego.plotting.utils_plotting.create_figure_axes(
        figure=figure, axes=axes
    )

    if bin_edges is None:
        bin_edges, symlog_linear_threshold = get_bin_edges(
            data=values,
            n_bins=n_bins,
            n_bins_linear=n_bins_linear,
            symlog_linear_threshold=symlog_linear_threshold,
            log=_log[0],
            return_symlog_linear_threshold=True,
            vmin=_xlim[0],
            vmax=_xlim[1],
        )
    (
        hist,
        bin_edges,
    ) = np.histogram(values, bins=bin_edges)
    bin_centers = bin_edges[:-1] + np.diff(bin_edges) / 2.0

    _ = plot_bar(
        bin_centers,
        hist,
        width_bars=np.diff(bin_edges),
        xlim=_xlim,  # type: ignore
        ylim=_ylim,  # type: ignore
        axes=axes,
        log=_log,  # type: ignore
        symlog_linear_threshold=symlog_linear_threshold,
        label_x=label_x,
        label_y=label_y,
        vertical=vertical,
        alpha=alpha,
        color=color,
        rotation_x_labels=rotation_x_labels,
        rotation_y_labels=rotation_y_labels,
        font_size=font_size,
    )
    if horizontal_line is not None:
        axes.axhline(horizontal_line)

    if tight_layout:
        figure.tight_layout()
    kego.plotting.utils_plotting.set_title(axes, title, font_size=font_size)
    kego.plotting.figures.save_figure(figure, filename)
    return axes


def plot_bar(
    bin_centers: np.ndarray,
    height_bars: np.ndarray,
    width_bars: np.ndarray,
    axes: kego.constants.TYPE_MATPLOTLIB_AXES,
    xlim: tuple[float | None, float | None] | None = None,
    ylim: tuple[float | None, float | None] | None = None,
    log: str | tuple[str, str] = "false",
    symlog_linear_threshold: None | float = None,
    label_x: None | str = None,
    label_y: None | str = None,
    vertical: bool = False,
    alpha: float = 1,
    color: str | None = None,
    font_size: float = kego.constants.DEFAULT_FONTSIZE_LARGE,
    replace_x_labels_at: Sequence | None = None,
    replace_x_labels_with: Sequence | None = None,
    replace_y_labels_at: Sequence | None = None,
    replace_y_labels_with: Sequence | None = None,
    rotation_x_labels: int = 0,
    rotation_y_labels: int = 0,
) -> matplotlib.container.BarContainer:
    """
    plots 1d bar plot

    Parameters:
    ----------
    bin_centers:
        Center position of bins, array(N)
    height_bars:
        Bar heights at corresponding `bin_centers`, array(N)
    width_bars:
        Bar widths, array(N)
    axes:
        Matplotlib axes
    xlim:
        Limits of x-axis and bin_edges, if None determined from (min, max) of x-values
    ylim:
        Limits of y-axis and bin_edges, if None determined from (min, max) of y-values
    log:
        Type of bins ("false", "symlog", "log")
    symlog_linear_threshold:
        Required when `log="symlog"`.
        Threshold value below which symlog becomes linear.
    label_x:
        Label of x-axis
    label_y:
        Label of y-axis
    vertical:
        Plot bars along vertical axis instead of default horizontal direction
    alpha:
        Alpha value of bars
    color:
        Color of bars
    font_size:
        Size of font
    replace_x_labels_at:
        Tick values of x-axis to be replaced
    replace_x_labels_with:
        Replacement values for tick values of x-axis
    replace_y_labels_at:
        Tick values of y-axis to be replaced
    replace_y_labels_with:
        Replacement values for tick values of y-axis
    rotation_x_labels:
        Angle of rotation of x-axis labels
    rotation_y_labels:
        Angle of rotation of y-axis labels

    Returns
    -------
    Bar plot object
    """
    _log = kego.lists.to_nlength_tuple(log)
    _xlim = kego.lists.to_nlength_tuple(xlim)
    _ylim = kego.lists.to_nlength_tuple(ylim)

    if vertical:
        plot = axes.barh(
            bin_centers,
            height_bars,
            height=width_bars,
            edgecolor="black",
            alpha=alpha,
            color=color,
        )
    else:
        plot = axes.bar(
            bin_centers,
            height_bars,
            width=width_bars,
            edgecolor="black",
            alpha=alpha,
            color=color,
        )

    kego.plotting.utils_plotting.set_x_log(
        axes, _log[0], axis_symlog_linear_threshold=symlog_linear_threshold
    )
    kego.plotting.utils_plotting.set_y_log(
        axes, _log[1], axis_symlog_linear_threshold=symlog_linear_threshold
    )
    kego.plotting.utils_plotting.set_axis_tick_labels(
        axes,
        replace_x_labels_at,
        replace_x_labels_with,
        axis="x",
        rotation=rotation_x_labels,
        font_size=font_size,
    )
    kego.plotting.utils_plotting.set_axis_tick_labels(
        axes,
        replace_y_labels_at,
        replace_y_labels_with,
        axis="y",
        rotation=rotation_y_labels,
        font_size=font_size,
    )
    kego.plotting.utils_plotting.set_axes_label(axes, label_x, "x", font_size=font_size)
    kego.plotting.utils_plotting.set_axes_label(axes, label_y, "y", font_size=font_size)
    axes.set_ylim(_ylim)
    axes.set_xlim(_xlim)
    return plot


def flatten_array(x):
    x = np.asarray(x)
    x = np.ndarray.flatten(x)
    return x


def get_bin_edges(
    vmin: float | None = None,
    vmax: float | None = None,
    data: np.ndarray | None = None,
    n_bins: int = 60,
    n_bins_linear: int = 10,
    log: str = "false",
    symlog_linear_threshold: float | None = None,
    return_symlog_linear_threshold: bool = False,
) -> tuple[np.ndarray, float | None] | np.ndarray:
    """
    Computes bin edges for plots

    Log types included `"false"`, `"symlog"`, `"log"`
    Parameters:
    ----------
    vmin:
        Minimum value of bin edge
    vmax:
        Maximum value of bin edge
    data:
        Used to compute `vmin`/`vmax` if None
    n_bins:
        Number of bins
    n_bins_linear:
        Only relevant if `log="symlog"`, Number of bins for linear part of bins (around zero)
    log:
        Type of bins ("false", "symlog", "log")
    symlog_linear_threshold:
        Required when `log="symlog"`.
        Threshold value below which symlog becomes linear.
    return_symlog_linear_threshold:
        Return symlog_linear_threshold in addition to bin edges

    Returns
    -------
    Bin edges, Optionally(symlog_linear_threshold)
    """
    if data is not None and vmin is None:
        vmin = data.min()
    if data is not None and vmax is None:
        vmax = data.max()
    if vmin is None or vmax is None:
        raise ValueError(
            f"Need to specify vmin {vmin} and {vmax} or provide data: {data}!"
        )
    if vmin > vmax:
        raise ValueError(f"{vmin=} > {vmax=}")
    if vmin <= 0 and log == "log":
        raise ValueError(f"For {log=}, cannot have {vmin=} <= 0")
    if vmin == vmax:
        vmin -= vmax / 2
        vmax += vmax / 2
        n_bins = 2
    if log == "symlog":
        if symlog_linear_threshold is None:
            abs_max = abs(vmax)
            abs_min = abs(vmin)
            symlog_linear_threshold = (
                abs_min if abs_min < abs_max else abs_max if abs_max != 0 else abs_min
            )
            if symlog_linear_threshold == 0:
                raise ValueError(
                    f"{symlog_linear_threshold=}, which is not allowed. Please set manually to different value!"
                )
            logger.info(
                f"Setting: linear_thresh: {symlog_linear_threshold} with vmin: {vmin}"
                " and vmax: {vmax}!"
            )

        bins = _get_bin_edges_symlog(
            vmin,
            vmax,
            symlog_linear_threshold,
            n_bins=n_bins,
            n_bins_linear=n_bins_linear,
        )
    elif log == "log":
        bins = 10 ** np.linspace(np.log10(vmin), np.log10(vmax), n_bins)
    else:
        bins = np.linspace(vmin, vmax, n_bins)

    if return_symlog_linear_threshold:
        return bins, symlog_linear_threshold
    else:
        return bins


def _get_bin_edges_symlog(
    vmin: float,
    vmax: float,
    linear_thresh: float,
    n_bins: int = 60,
    n_bins_linear: int = 10,
) -> np.ndarray:
    """
    Computes symmetrical logarithmic bins

    Bins have same absolute vmin/vmax if vmin is negative
    Parameters:
    ----------
    vmin:
        Minimum value of bin edge
    vmax:
        Maximum value of bin edge
    linear_thresh:
        Threshold value below which symlog becomes linear.
    n_bins:
        Number of bins for logarithmic part of bins
    n_bins_linear:
        Number of bins for linear part of bins (around zero)

    Returns
    -------
    symmetrical bin edges
    """
    if isinstance(vmin, np.datetime64) or vmin > 0:
        bins = 10 ** np.linspace(np.log10(vmin), np.log10(vmax), n_bins)
    elif vmin == 0:
        bins = np.hstack(
            (
                np.linspace(0, linear_thresh, n_bins_linear),
                10 ** np.linspace(np.log10(linear_thresh), np.log10(vmax), n_bins),
            )
        )
    else:
        bins = np.hstack(
            (
                -(
                    10
                    ** np.linspace(
                        np.log10(vmax),
                        np.log10(linear_thresh),
                        n_bins // 2,
                        endpoint=False,
                    )
                ),
                np.linspace(
                    -linear_thresh, linear_thresh, n_bins_linear, endpoint=False
                ),
                10 ** np.linspace(np.log10(linear_thresh), np.log10(vmax), n_bins // 2),
            )
        )
    return bins


def _find_axis_limits(
    xlim: tuple[float | None, float | None],
    ylim: tuple[float | None, float | None],
    bin_edges_x: np.ndarray,
    bin_edges_y: np.ndarray,
) -> tuple[tuple[float, float], tuple[float, float]]:
    if xlim[0] is None:
        xlim = (bin_edges_x[0], xlim[1])
    if xlim[1] is None:
        xlim = (xlim[0], bin_edges_x[-1])
    if ylim[0] is None:
        ylim = (bin_edges_y[0], ylim[1])
    if ylim[1] is None:
        ylim = (ylim[0], bin_edges_y[-1])
    return xlim, ylim  # type: ignore



from typing import Literal

import numpy as np

from kego.constants import DEFAULT_COLORMAP
from kego.lists import to_nlength_tuple
from kego.plotting.axes_utils import set_axes
from kego.plotting.colormesh import _plot_colormesh
from kego.plotting.utils_plotting import create_axes_grid, get_norm


def grid_plot(
    data_as_dict: dict,
    xs: list | None = None,
    ys: list | None = None,
    style: Literal["colormesh"] = "colormesh",
    labels: list | None = None,
    nx_max: int = 3,
    vmin: float | None = None,
    vmax: float | None = None,
    colormap: str = DEFAULT_COLORMAP,
):
    if data_as_dict is None:
        if xs is None or ys is None:
            raise ValueError(
                f"{data_as_dict=} is not set. Please specify both {xs=} and {ys=} instead"
            )
    if data_as_dict is not None and (xs is not None or ys is not None):
        raise ValueError(f"Can only specify {data_as_dict=} or [{xs=} and {ys=}].")
    if xs is not None or ys is not None:
        raise NotImplementedError(f"Specifying {xs=} or {ys=} is not implemented!")

    n_plots = len(data_as_dict.items())
    n_columns = nx_max if n_plots >= nx_max else n_plots
    n_rows = int(np.ceil(n_plots / nx_max))
    figure, axes_grid, axes_colorbar = create_axes_grid(
        n_columns=n_columns, n_rows=n_rows
    )

    for i_plot, (label, data) in enumerate(data_as_dict.items()):
        if style == "colormesh":
            axes = np.ndarray.flatten(axes_grid)[i_plot]
            xx, yy = np.meshgrid(
                np.arange(np.shape(data)[1]), np.arange(np.shape(data)[0])
            )
            print(f"{xx.shape=}, {yy.shape=}, {data.shape=}")
            norm = get_norm("linear", vmin=vmin, vmax=vmax)
            _plot_colormesh(
                axes=axes,
                xx=xx,
                yy=yy,
                matrix=data,
                norm_object=norm,
                colormap=colormap,
            )
            set_axes(ax=axes, title=label)
    return figure



from typing import Literal

import numpy as np
import pandas as pd

import kego.checks
import kego.constants


def _resolve_defaults(
    facet_column: str | None,
    facet_row: str | None,
    marginal_x: str | None,
    marginal_y: str | None,
    spacing_x: float | None,
    spacing_y: float | None,
    axes_style: kego.constants.AXES_STYLE_TYPE,
) -> tuple[bool, None | str, None | str, float, float]:
    if axes_style == "axes_single" and (
        facet_column is not None or facet_row is not None
    ):
        raise ValueError(
            f"Cannot specify {facet_column=} or {facet_row=} when axes is given!"
        )
    if marginal_x == "default":
        if facet_column is None and facet_row is None and axes_style == "axes_grid":
            marginal_x = "histogram"
        else:
            marginal_x = None
    if marginal_y == "default":
        if facet_column is None and facet_row is None and axes_style == "axes_grid":
            marginal_y = "histogram"
        else:
            marginal_y = None
    no_marginal_plots = marginal_x is None and marginal_y is None
    if facet_column is not None and spacing_x is None:
        spacing_x = 0.1
    else:
        spacing_x = 0.03
    if facet_row is not None and spacing_y is None:
        spacing_y = 0.08
    else:
        spacing_y = 0.03
    return no_marginal_plots, marginal_x, marginal_y, spacing_x, spacing_y


def _resolve_axes_style(
    axes: kego.constants.TYPE_MATPLOTLIB_AXES | None,
) -> kego.constants.AXES_STYLE_TYPE:
    axes_style: Literal["axes_single", "axes_grid"]
    if axes is not None:
        axes_style = "axes_single"
    else:
        axes_style = "axes_grid"
    return axes_style


def _check_parameter_consistency(
    facet_column: str | None,
    facet_row: str | None,
    marginal_x: str | None,
    marginal_y: str | None,
    df: pd.DataFrame | None,
):
    if (bool(marginal_x) + bool(marginal_y)) not in [0, 2]:
        raise NotImplementedError(
            f"{bool(marginal_x)=} anbool(d) {marginal_y=} both have to be on/off"
        )
    if (bool(marginal_x) or bool(marginal_y)) and (facet_column or facet_row):
        raise NotImplementedError(
            f"Cannot use ({marginal_x=} or {marginal_y}) at the same time as ({facet_column=} or {facet_row})"
        )
    if (facet_column or facet_row) and df is None:
        raise NotImplementedError(
            f"Cannnot use {facet_column=} and/or {facet_row}, when {df=} is not given!"
        )


def _prepare_facet(
    df: pd.DataFrame | None, facet: str | None
) -> tuple[np.ndarray | None, int, np.ndarray | None]:
    if facet is not None and isinstance(df, pd.DataFrame) and facet in df:
        facet_values = np.unique(df[facet].values)
        masks = np.array([df[facet] == val for val in facet_values])
        n_facet_values = len(facet_values)
        facet_title = np.array([f"{facet} == {val}" for val in facet_values])
    else:
        facet_values, masks, n_facet_values, facet_title = None, None, 0, None
    return masks, n_facet_values, facet_title


def _set_number_rows_and_columns(
    marginal_x: str | None,
    marginal_y: str | None,
    n_facet_column_values: int | None,
    n_facet_row_values: int | None,
) -> tuple[int, int]:
    n_columns = 1
    if n_facet_column_values:
        n_columns = n_facet_column_values
    elif marginal_x is not None:
        n_columns = 2

    n_rows = 1
    if n_facet_row_values:
        n_rows = n_facet_row_values
    elif marginal_y is not None:
        n_rows = 2
    return n_columns, n_rows


def _set_axis_labels(
    n_columns: int,
    n_rows: int,
    facet_show_label_x: list[int] | str = "minimal",
    facet_show_label_y: list[int] | str = "minimal",
    label_x: str | None = None,
    label_y: str | None = None,
    facet_column: str | None = None,
    facet_row: str | None = None,
) -> tuple[np.ndarray, np.ndarray]:
    if facet_column is not None:
        i_last_row = n_rows - 1
        xaxis_labels = np.full((n_rows, n_columns), "off", dtype=object)
        for i_row in range(n_rows):
            for i_column in range(n_columns):
                if isinstance(facet_show_label_x, str):
                    if facet_show_label_x == "all":
                        xaxis_labels[i_row, i_column] = label_x
                    elif facet_show_label_x == "minimal":
                        if i_row == i_last_row:
                            xaxis_labels[i_row, i_column] = label_x
                else:
                    kego.checks.assert_all_same_type(facet_show_label_x, int)
                    if i_row in facet_show_label_x:
                        xaxis_labels[i_row, i_column] = label_x
    else:
        xaxis_labels = np.full((n_rows, n_columns), label_x, dtype=object)
    if facet_row is not None:
        i_first_colum = 0
        yaxis_labels = np.full((n_rows, n_columns), "off", dtype=object)
        for i_row in range(n_rows):
            for i_column in range(n_columns):
                if isinstance(facet_show_label_y, str):
                    if facet_show_label_y == "all":
                        yaxis_labels[i_row, i_column] = label_y
                    elif facet_show_label_y == "minimal":
                        if i_column == i_first_colum:
                            yaxis_labels[i_row, i_column] = label_y
                else:
                    kego.checks.assert_all_same_type(facet_show_label_y, int)
                    if i_column in facet_show_label_y:
                        yaxis_labels[i_row, i_column] = label_y
    else:
        yaxis_labels = np.full((n_rows, n_columns), label_y, dtype=object)
    return xaxis_labels, yaxis_labels


def _set_masks(
    column_masks: np.ndarray | None,
    row_masks: np.ndarray | None,
    n_columns: int,
    n_rows: int,
) -> np.ndarray:
    masks = np.full((n_rows, n_columns), None, dtype=object)
    for i_row in range(n_rows):
        for i_column in range(n_columns):
            if column_masks is not None:
                masks[i_row, i_column] = _add_masks(
                    masks[i_row, i_column], column_masks[i_column]
                )
            if row_masks is not None:
                masks[i_row, i_column] = _add_masks(
                    masks[i_row, i_column], row_masks[i_row]
                )
    return masks


def _determine_axes_titles(
    facet_column: str | None,
    facet_row: str | None,
    title: str,
    facet_column_title: np.ndarray | None,
    facet_row_title: np.ndarray | None,
    n_columns: int,
    n_rows: int,
) -> np.ndarray:
    titles = np.full((n_rows, n_columns), title, dtype=object)
    for i_row in range(n_rows):
        for i_column in range(n_columns):
            if facet_row and facet_row_title is not None:
                titles[i_row, i_column] = _add_titles(
                    [titles[i_row, i_column], facet_row_title[i_row]]
                )
            if facet_column and facet_column_title is not None:
                titles[i_row, i_column] = _add_titles(
                    [titles[i_row, i_column], facet_column_title[i_column]]
                )
    return titles


def _prepare_axes_grid_creation(
    marginal_x: str | None, marginal_y: str | None
) -> tuple[
    list[tuple[int, int]] | None,
    list[float] | None,
    list[float] | None,
    list[tuple[int, int]] | None,
    bool,
]:
    skip_row_col: list[tuple[int, int]] | None = None
    widths_along_x: list[float] | None = None
    heights_along_y: list[float] | None = None
    colorbar_include_row_col: list[tuple[int, int]] | None = None

    colorbar_off = False
    if marginal_x is not None and marginal_y is not None:
        skip_row_col = [
            (0, 1),
        ]
        widths_along_x = [0.2, 0.1]
        heights_along_y = [0.1, 0.2]
        colorbar_include_row_col = [
            (1, 1),
        ]
    return (
        skip_row_col,
        widths_along_x,
        heights_along_y,
        colorbar_include_row_col,
        colorbar_off,
    )


def _add_titles(titles: list, delimeter: str = ", "):
    return f"{delimeter}".join([t for t in titles if len(t)])


def _add_masks(mask: np.ndarray | None, to_add: np.ndarray | None) -> np.ndarray | None:
    if to_add is None:
        return mask
    if mask is None:
        return to_add
    else:
        return mask & to_add



import numpy as np
import pandas as pd
import pandas.api.types
import scipy.stats


class ParticipantVisibleError(Exception):
    pass


def score(
    solution: pd.DataFrame,
    submission: pd.DataFrame,
    row_id_column_name: str | None,
    naive_mean: float,
    naive_sigma: float,
    sigma_true: float,
) -> float:
    """
    This is a Gaussian Log Likelihood based metric. For a submission, which contains the predicted mean (x_hat) and variance (x_hat_std),
    we calculate the Gaussian Log-likelihood (GLL) value to the provided ground truth (x). We treat each pair of x_hat,
    x_hat_std as a 1D gaussian, meaning there will be 283 1D gaussian distributions, hence 283 values for each test spectrum,
    the GLL value for one spectrum is the sum of all of them.

    Inputs:
        - solution: Ground Truth spectra (from test set)
            - shape: (nsamples, n_wavelengths)
        - submission: Predicted spectra and errors (from participants)
            - shape: (nsamples, n_wavelengths*2)
        naive_mean: (float) mean from the train set.
        naive_sigma: (float) standard deviation from the train set.
        sigma_true: (float) essentially sets the scale of the outputs.
    """
    solution = solution.copy()
    submission = submission.copy()
    if row_id_column_name is not None:
        del solution[row_id_column_name]
        del submission[row_id_column_name]

    if submission.min().min() < 0:
        raise ParticipantVisibleError("Negative values in the submission")
    for col in submission.columns:
        if not pandas.api.types.is_numeric_dtype(submission[col]):
            raise ParticipantVisibleError(f"Submission column {col} must be a number")

    n_wavelengths = len(solution.columns)
    if len(submission.columns) != n_wavelengths * 2:
        raise ParticipantVisibleError("Wrong number of columns in the submission")

    y_pred = submission.iloc[:, :n_wavelengths].values
    # Set a non-zero minimum sigma pred to prevent division by zero errors.
    sigma_pred = np.clip(
        submission.iloc[:, n_wavelengths:].values, a_min=10**-15, a_max=None
    )
    y_true = solution.values

    GLL_pred = np.sum(scipy.stats.norm.logpdf(y_true, loc=y_pred, scale=sigma_pred))
    GLL_true = np.sum(
        scipy.stats.norm.logpdf(
            y_true, loc=y_true, scale=sigma_true * np.ones_like(y_true)
        )
    )
    GLL_mean = np.sum(
        scipy.stats.norm.logpdf(
            y_true,
            loc=naive_mean * np.ones_like(y_true),
            scale=naive_sigma * np.ones_like(y_true),
        )
    )

    submit_score = (GLL_pred - GLL_mean) / (GLL_true - GLL_mean)
    return float(np.clip(submit_score, 0.0, 1.0))


def compute_metrics(solution, submission_values, submission_sigma):
    submission = np.concat((submission_values, submission_sigma), axis=1)
    return score(
        solution=solution,
        submission=submission,
        row_id_column_name="planet_id",
        naive_mean=np.mean(submission_values),
        naive_sigma=np.std(submission_values),
        sigma_true=np.std(solution),
    )



import os

import numpy as np

from kego.constants import TYPE_FILEPATHS


def list_files(
    path,
    match_filename: str | None = None,
    level_max: int = np.inf,
    return_absolute_path: bool = False,
):
    path = path.__str__()
    c = path.count(os.sep)
    for root, dirs, files in os.walk(path):
        for name in files:
            result = os.path.join(root, name)
            if root.count(os.sep) - c - 1 <= level_max and (
                match_filename is None or name == match_filename
            ):
                if return_absolute_path:
                    result = os.path.abspath(result)
                yield result



import pathlib
from typing import Literal, TypeAlias

import matplotlib.axes
import matplotlib.figure
import pandas as pd

TIME_TYPE_PANDAS = pd.Timestamp
TYPE_MATPLOTLIB_AXES: TypeAlias = matplotlib.axes.Axes
TYPE_MATPLOTLIB_FIGURES: TypeAlias = matplotlib.figure.Figure
TYPE_MATPLOTLIB_COLORBAR: TypeAlias = matplotlib.colorbar.Colorbar
TYPE_FILEPATHS: TypeAlias = str | pathlib.Path
AXES_STYLE_TYPE: TypeAlias = Literal["axes_single", "axes_grid"]
DEFAULT_COLORMAP: str = "viridis"
DEFAULT_FONTSIZE_LARGE: float = 20
DEFAULT_FONTSIZE_SMALL: float = 12
DEFAULT_FIGURE_SIZE: tuple[float, float] = (10.0, 6.0)
